#include "QuadTreeRenderer.h"
#include "glFuncs.h"
#include "fastmath.h"

GEZERROR QuadTreeRenderer::InitImmediateMode(Preferences &prefs)
{
	int error;

	error = 0;

	QuadTreeRender = &QuadTreeRenderer::QuadTreeRender_Immediate;

	if (prefs.GET_INT_PARAM(PARAM_USE_VERTEX_ARRAY))
	{
		console->Output("enabling vertex arrays\n");

		if (index_data != NULL)
		{
			delete[] index_data;
		}

		index_data = new GLuint[MAP_WIDTH * MAP_HEIGHT * 12]; // assume worst case
		ASSERT(index_data != NULL, "out of memory");
		if (index_data == NULL)
		{
			return GEZ_UNKNOWN;
		}

		if (multi_indexes != NULL) delete[] multi_indexes;
		multi_indexes = new GLuint*[MAP_WIDTH * MAP_HEIGHT];
		ASSERT(index_data != NULL, "out of memory");
		if (index_data == NULL)
		{
			return GEZ_UNKNOWN; // TODO: don't fail, we could fall back...
		} // TODO: don't do that if not needed

		if (multi_lengths != NULL) delete[] multi_lengths;
		multi_lengths = new GLsizei[MAP_WIDTH * MAP_HEIGHT];
		ASSERT(multi_lengths != NULL, "out of memory");
		if (multi_lengths == NULL)
		{
			return GEZ_UNKNOWN;
		}

		if (prefs.GET_INT_PARAM(PARAM_USE_MULTIDRAW_EXT))
		{
			console->Output("querying GL_EXT_multi_draw_arrays ");
			if (QueryExtension("GL_EXT_multi_draw_arrays"))
			{
				console->Output("\n");

				error = 0;

				GL_QUERY_PROC(glMultiDrawArraysEXT,   PFNGLMULTIDRAWARRAYSEXTPROC)
				GL_QUERY_PROC(glMultiDrawElementsEXT, PFNGLMULTIDRAWELEMENTSEXTPROC);

				if (!error)
				{
					multi_draw = true;
				}
				else
					multi_draw = false;
			}
			else
			{
				multi_draw = false;

				console->Output("not available\n");
			}
		}

		QuadTreeRender = (QUADTREERENDER) &QuadTreeRenderer::QuadTreeRender_Immediate_VertexArray;

		glVertexPointer(3, GL_FLOAT, 0, vertex_data);
		glColorPointer(3, GL_FLOAT, 0, color_data);
		glTexCoordPointer(2, GL_FLOAT, 0, texture_coords);

		glEnable(GL_VERTEX_ARRAY);
		glEnable(GL_COLOR_ARRAY);
		glEnable(GL_TEXTURE_COORD_ARRAY);

		if (prefs.GET_INT_PARAM(PARAM_USE_COMPILED_VERTEX_ARRAY))
		{
			console->Output("querying GL_EXT_compiled_vertex_array ");
			if (QueryExtension("GL_EXT_compiled_vertex_array"))
			{
				console->Output("\n");

				error = 0;
				GL_QUERY_PROC(glLockArraysEXT, PFNGLLOCKARRAYSEXTPROC);
				GL_QUERY_PROC(glUnlockArraysEXT, PFNGLUNLOCKARRAYSEXTPROC);

				if (!error)
				{
					glLockArraysEXT(0, MAP_WIDTH * MAP_HEIGHT);
					ASSERT(glGetError() == GL_NO_ERROR, "error");

					compiled_array = true;
				}

			}
			else
			{
				console->Output("not available\n");
			}
		}
	}

	return GEZ_OK;
}
void QuadTreeRenderer::QuadTreeRender_Immediate_VertexArray(unsigned int x1, unsigned int y1, unsigned int x2, unsigned int y2)
{
	register GLuint i;

	index_count = 0;
	fan_current = 0;

	index_pointer = index_data;
	QuadTreeRender_TriangleFan_Recursion(x1, y1, x2, y2);

	if (multi_draw)
	{
		glMultiDrawElementsEXT(GL_TRIANGLE_FAN, multi_lengths, GL_UNSIGNED_INT, (const GLvoid **)multi_indexes, fan_current);
	}
	else
	{

		for (i = 0; i < fan_current; i++)
		{
			glDrawElements(GL_TRIANGLE_FAN, multi_lengths[i], GL_UNSIGNED_INT, multi_indexes[i]);
		}
	}

	vertices_count = index_count;
}

void QuadTreeRenderer::QuadTreeRender_TriangleFan_Recursion(unsigned int x1, unsigned int y1, unsigned int x2, unsigned int y2)
{
	int x, y, edge_len_x, edge_len_y;
	unsigned int MAPWIDTH_times_y,
		MAPWIDTH_times_y1,
		MAPWIDTH_times_y2;

	ASSERT(multi_indexes != NULL, "oops");
	ASSERT(multi_lengths != NULL, "oops, I did it again");

	edge_len_x = (x2 - x1);
	edge_len_y = (y2 - y1); // differ because of integer arithmetics

	x = x1 + edge_len_x / 2; 
	y = y1 + edge_len_y / 2;

	MAPWIDTH_times_y = MAP_WIDTH * y;

	if ((quadtree_data[MAPWIDTH_times_y + x] == NODE_REFINE) && (edge_len_x > 3))
	{
		QuadTreeRender_TriangleFan_Recursion(x1,  y1,  x,  y);
		QuadTreeRender_TriangleFan_Recursion( x,  y1, x2,  y);
		QuadTreeRender_TriangleFan_Recursion( x,  y,  x2, y2);
		QuadTreeRender_TriangleFan_Recursion(x1,  y,  x,  y2);
	}
	else
	{
		if (quadtree_data[MAPWIDTH_times_y + x] == NODE_RENDER)
		{
			MAPWIDTH_times_y1 = MAP_WIDTH * y1;
			MAPWIDTH_times_y2 = MAP_WIDTH * y2;

			GLuint dummy;
			dummy = index_count;

			ASSERT(index_count < MAP_WIDTH * MAP_HEIGHT, "too many vertices");

			multi_indexes[fan_current] = &index_pointer[index_count];
			index_pointer[index_count++] = MAPWIDTH_times_y + x;
			index_pointer[index_count++] = MAP_WIDTH * y2 + x2; // right bottom 

			// test whether right node is coarser 
			if (x + edge_len_x < MAP_WIDTH)
			{
				if (quadtree_data[MAPWIDTH_times_y + x + edge_len_x] > 0 )
				{
					index_pointer[index_count++] = MAPWIDTH_times_y + x2; // right 
				}
			}
			else
			{
				index_pointer[index_count++] = MAPWIDTH_times_y + x2; // right 
			}
	        
			index_pointer[index_count++] = MAPWIDTH_times_y1 + x2; // right top 

			// prevent cracking with upper neighbour 
			if (y - edge_len_y >= 0)
			{
				if (quadtree_data[MAP_WIDTH * (y - edge_len_y) + x] > 0 )
				{
					index_pointer[index_count++] = MAPWIDTH_times_y1 +  x; // top 
				}
			}
			else
			{
				index_pointer[index_count++] = MAPWIDTH_times_y1 +  x; // top 
			}

			index_pointer[index_count++] = MAPWIDTH_times_y1 + x1; // left top 

			// prevent cracking with left neighbour
			if (x - edge_len_x >= 0)
			{
				if (quadtree_data[MAPWIDTH_times_y + x - edge_len_x] > 0 )
				{
					index_pointer[index_count++] = MAPWIDTH_times_y + x1; // left 
				}
			}
			else
			{
				index_pointer[index_count++] = MAPWIDTH_times_y + x1; // left 
			}

			index_pointer[index_count++] = MAPWIDTH_times_y2 + x1; // left bottom 
			
			// prevent cracking with lower neighbour 
			if (y + edge_len_y < MAP_HEIGHT)
			{
				if (quadtree_data[MAP_WIDTH * (y + edge_len_y) + x] > 0 )
				{
					index_pointer[index_count++] = MAPWIDTH_times_y2 +  x; // bottom 
				}
			}
			else
			{
				index_pointer[index_count++] = MAPWIDTH_times_y2 +  x; // bottom 
			}

			index_pointer[index_count++] = MAPWIDTH_times_y2 + x2; // right bottom 

			multi_lengths[fan_current] = index_count - dummy;
			++fan_current;

			triangles_count += index_count - dummy - 2;
		}
	}
}

// most basic rendering algorithm of them all

// TODO: don't call glTexture if there's nothing ter do
void QuadTreeRenderer::QuadTreeRender_Immediate(unsigned int x1, unsigned int y1, unsigned int x2, unsigned int y2)
{
	int x, y, edge_len_x, edge_len_y;
	int MAPWIDTH_times_y,
		MAPWIDTH_times_y1,
		MAPWIDTH_times_y2;

	edge_len_x = (x2 - x1);
	edge_len_y = (y2 - y1); // differ because of integer arithmetics

	x = x1 + (edge_len_x >> 1); 
	y = y1 + (edge_len_y >> 1);

	MAPWIDTH_times_y = MAP_WIDTH * y;

	if ((quadtree_data[MAP_HEIGHT * y + x] == NODE_REFINE) && (edge_len_x > 3))
	{
		QuadTreeRender_Immediate(x1,  y1,  x,  y);
		QuadTreeRender_Immediate( x,  y1, x2,  y);
		QuadTreeRender_Immediate( x,  y,  x2, y2);
		QuadTreeRender_Immediate(x1,  y,  x,  y2);
	}
	else
	{
		// render
		if (quadtree_data[MAP_HEIGHT * y + x] == NODE_RENDER)
		{
			int v_count;

			v_count = 0;
			MAPWIDTH_times_y1 = MAP_WIDTH * y1;
			MAPWIDTH_times_y2 = MAP_WIDTH * y2;

			glBegin(GL_TRIANGLE_FAN);

				glColor3f(color_data[MAPWIDTH_times_y +  x][0], color_data[MAPWIDTH_times_y +  x][0], color_data[MAPWIDTH_times_y +  x][0]);
				glTexCoord2f(texture_coords[MAPWIDTH_times_y + x][0], texture_coords[MAPWIDTH_times_y + x][1]);
				glVertex3f( x, vertex_data[(MAPWIDTH_times_y + x)][1],  y);
				++v_count;

		        glColor3f(color_data[MAPWIDTH_times_y2 +  x2][0], color_data[MAPWIDTH_times_y2 +  x2][0], color_data[MAPWIDTH_times_y2 +  x2][0]);
				glTexCoord2f(texture_coords[MAPWIDTH_times_y2 + x2][0], texture_coords[MAPWIDTH_times_y2 + x2][1]);
		        glVertex3f(x2, vertex_data[(MAPWIDTH_times_y2 + x2 )][1], y2); // right bottom 
				++v_count;

				// test ob right daneben coarser 
				if (x + edge_len_x < MAP_WIDTH)
				{
					if (quadtree_data[MAPWIDTH_times_y + x + edge_len_x] > 0 )
					{
						glColor3f(color_data[MAPWIDTH_times_y +  x2][0], color_data[MAPWIDTH_times_y +  x2][0], color_data[MAPWIDTH_times_y +  x2][0]);
						glTexCoord2f(texture_coords[MAPWIDTH_times_y + x2][0], texture_coords[MAPWIDTH_times_y + x2][1]);
						glVertex3f(x2, vertex_data[(MAPWIDTH_times_y + x2 )][1],  y); // right
						++v_count;
					}
				}
				else
				{
					glColor3f(color_data[MAPWIDTH_times_y +  x2][0], color_data[MAPWIDTH_times_y +  x2][0], color_data[MAPWIDTH_times_y +  x2][0]);
					glTexCoord2f(texture_coords[MAPWIDTH_times_y + x2][0], texture_coords[MAPWIDTH_times_y + x2][1]);
					glVertex3f(x2, vertex_data[(MAPWIDTH_times_y + x2 )][1],  y); // right
					++v_count;
				}
		        
				glColor3f(color_data[MAPWIDTH_times_y1 +  x2][0], color_data[MAPWIDTH_times_y1 +  x2][0], color_data[MAPWIDTH_times_y1 +  x2][0]);
				glTexCoord2f(texture_coords[MAPWIDTH_times_y1 + x2][0], texture_coords[MAPWIDTH_times_y1 + x2][1]);
				glVertex3f(x2, vertex_data[(MAPWIDTH_times_y1 + x2 )][1], y1); // right top 
				++v_count;

				// prevent cracking with upper neighbour 
				if (y - edge_len_y >= 0)
				{
					if (quadtree_data[MAP_WIDTH * (y - edge_len_y) + x] > 0 )
					{
						glColor3f(color_data[MAPWIDTH_times_y1 +  x][0], color_data[MAPWIDTH_times_y1 +  x][0], color_data[MAPWIDTH_times_y1 +  x][0]);
						glTexCoord2f(texture_coords[MAPWIDTH_times_y1 + x][0], texture_coords[MAPWIDTH_times_y1 + x][1]);
						glVertex3f( x, vertex_data[(MAPWIDTH_times_y1 +  x )][1], y1); // top 
						++v_count;
					}
				}
				else
				{
					glColor3f(color_data[MAPWIDTH_times_y1 +  x][0], color_data[MAPWIDTH_times_y1 +  x][0], color_data[MAPWIDTH_times_y1 +  x][0]);
					glTexCoord2f(texture_coords[MAPWIDTH_times_y1 + x][0], texture_coords[MAPWIDTH_times_y1 + x][1]);
					glVertex3f( x, vertex_data[(MAPWIDTH_times_y1 + x)][1], y1); // top 
					++v_count;
				}

				glColor3f(color_data[MAPWIDTH_times_y1 +  x1][0], color_data[MAPWIDTH_times_y1 +  x1][0], color_data[MAPWIDTH_times_y1 +  x1][0]);
				glTexCoord2f(texture_coords[MAPWIDTH_times_y1 + x1][0], texture_coords[MAPWIDTH_times_y1 + x1][1]);
				glVertex3f(x1, vertex_data[(MAPWIDTH_times_y1 + x1)][1], y1); // left top 
				++v_count;

				// prevent cracking with left neighbour
				if (x - edge_len_x >= 0)
				{
					if (quadtree_data[MAPWIDTH_times_y + x - edge_len_x] > 0 )
				{
					glColor3f(color_data[MAPWIDTH_times_y +  x1][0], color_data[MAPWIDTH_times_y +  x1][0], color_data[MAPWIDTH_times_y +  x1][0]);
					glTexCoord2f(texture_coords[MAPWIDTH_times_y + x1][0], texture_coords[MAPWIDTH_times_y + x1][1]);
					glVertex3f(x1, vertex_data[(MAPWIDTH_times_y + x1 )][1],  y); // left 
					++v_count;
				}
			}
			else
			{
				glColor3f(color_data[MAPWIDTH_times_y +  x1][0], color_data[MAPWIDTH_times_y +  x1][0], color_data[MAPWIDTH_times_y +  x1][0]);
				glTexCoord2f(texture_coords[MAPWIDTH_times_y + x1][0], texture_coords[MAPWIDTH_times_y + x1][1]);
				glVertex3f(x1, vertex_data[(MAPWIDTH_times_y + x1)][1],  y); // left 
				++v_count;
			}

			glColor3f(color_data[MAPWIDTH_times_y2 +  x1][0], color_data[MAPWIDTH_times_y2 +  x1][0], color_data[MAPWIDTH_times_y2 +  x1][0]);
			glTexCoord2f(texture_coords[MAPWIDTH_times_y2 + x1][0], texture_coords[MAPWIDTH_times_y2 + x1][1]);
			glVertex3f(x1, vertex_data[(MAPWIDTH_times_y2 + x1 )][1], y2); // left bottom 
			++v_count;
			
			// prevent cracking with lower neighbour 
			if (y + edge_len_y < MAP_HEIGHT)
			{
				if (quadtree_data[MAP_WIDTH * (y + edge_len_y) + x] > 0 )
				{
					glColor3f(color_data[MAPWIDTH_times_y2 +  x][0], color_data[MAPWIDTH_times_y2 +  x][0], color_data[MAPWIDTH_times_y2 +  x][0]);
					glTexCoord2f(texture_coords[MAPWIDTH_times_y2 + x][0], texture_coords[MAPWIDTH_times_y2 + x][1]);
					glVertex3f( x, vertex_data[(MAPWIDTH_times_y2 +  x )][1], y2); // bottom 
					++v_count;
				}
			}
			else
			{
				glColor3f(color_data[MAPWIDTH_times_y2 +  x][0], color_data[MAPWIDTH_times_y2 +  x][0], color_data[MAPWIDTH_times_y2 +  x][0]);
				glTexCoord2f(texture_coords[MAPWIDTH_times_y2 + x][0], texture_coords[MAPWIDTH_times_y2 + x][1]);
				glVertex3f( x, vertex_data[(MAPWIDTH_times_y2 +  x )][1], y2); // bottom 
				++v_count;
			}

			glColor3f(color_data[MAPWIDTH_times_y2 +  x2][0], color_data[MAPWIDTH_times_y2 +  x2][0], color_data[MAPWIDTH_times_y2 + x2][0]);
			glTexCoord2f(texture_coords[MAPWIDTH_times_y2 + x2][0], texture_coords[MAPWIDTH_times_y2 + x2][1]);
			glVertex3f(x2, vertex_data[(MAPWIDTH_times_y2 + x2 )][1], y2); // right bottom 
			++v_count;

		glEnd();

		triangles_count += v_count - 2;
		vertices_count += v_count;
	}

#ifdef DRAW_CULLED_NODES
		else if (quadtree_data[MAP_HEIGHT * y + x] == NODE_CULLED)
		{
			      glBegin(GL_TRIANGLE_FAN);

				  glColor3f(1.0f, 0.0, 0.0f);
        glVertex3f( x, vertex_data[(MAP_WIDTH *  y +  x )][1],  y);

        glVertex3f(x2, vertex_data[(MAP_WIDTH * y2 + x2 )][1], y2); // right bottom 

		// test whether right node is coarser 
		if (x + edge_len_x < MAP_WIDTH)
		{
			if (quadtree_data[MAP_WIDTH * y + x + edge_len_x] > 0 )
			{
		        glVertex3f(x2, vertex_data[(MAP_WIDTH *  y + x2 )][1],  y); // right 
			}
		}
		else
		{
	        glVertex3f(x2, vertex_data[(MAP_WIDTH *  y + x2 )][1],  y); // right 
		}
        
		glVertex3f(x2, vertex_data[(MAP_WIDTH * y1 + x2 )][1], y1); // right top 

        // prevent cracking with upper neighbour 
		if (y - edge_len_y >= 0)
		{
			if (quadtree_data[MAP_WIDTH * (y - edge_len_y) + x] > 0 )
			{
				glVertex3f( x, vertex_data[(MAP_WIDTH * y1 +  x )][1], y1); // top 
			}
		}
		else
		{
			glVertex3f( x, vertex_data[(MAP_WIDTH * y1 +  x )][1], y1); // top 
		}

		glVertex3f(x1, vertex_data[(MAP_WIDTH * y1 + x1 )][1], y1); // left top 

		// prevent cracking with left neighbour
		if (x - edge_len_x >= 0)
		{
			if (quadtree_data[MAP_WIDTH * y + x - edge_len_x] > 0 )
			{
		        glVertex3f(x1, vertex_data[(MAP_WIDTH *  y + x1 )][1],  y); // left 
			}
		}
		else
		{
			glVertex3f(x1, vertex_data[(MAP_WIDTH *  y + x1 )][1],  y); // left 
		}

        glVertex3f(x1, vertex_data[(MAP_WIDTH * y2 + x1 )][1], y2); // left bottom 
		
        // prevent cracking with lower neighbour 
		if (y + edge_len_y < MAP_HEIGHT)
		{
			if (quadtree_data[MAP_WIDTH * (y + edge_len_y) + x] > 0 )
			{
				glVertex3f( x, vertex_data[(MAP_WIDTH * y2 +  x )][1], y2); // bottom 
			}
		}
		else
		{
				glVertex3f( x, vertex_data[(MAP_WIDTH * y2 +  x )][1], y2); // bottom 
		}

        glVertex3f(x2, vertex_data[(MAP_WIDTH * y2 + x2 )][1], y2); // right bottom 
      glEnd();

		}
#endif // #defined DRAW_CULLED_NODES
	}
}
