#include "culling.h"
#include "Assert.h"

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <math.h>
#include <GL/gl.h>

Halfspace __forceinline ClassifyPoint(const Plane &plane, const Point &pt)
{
	float d;
	d = plane.a * pt.x + plane.b * pt.y + plane.c * pt.z + plane.d;
	if (d < 0) return NEGATIVE;
	if (d > 0) return POSITIVE;
	return ON_PLANE;
}

Visibility checkPlaneSide(const Plane &plane, const Point p[4])
{
	if (ClassifyPoint(plane, p[0]) == NEGATIVE)
	{
		// near left is left of left clipping plane
		if (ClassifyPoint(plane, p[1]) == NEGATIVE)
		{
			// near right is left of left clipping plane
			if (ClassifyPoint(plane, p[2]) == NEGATIVE)
			{
				// far right is left of left clipping plane
				if (ClassifyPoint(plane, p[3]) == NEGATIVE)
				{
					// far left is left of left clipping plane
					return NONE;
				}
			}
		}

		return PARTIAL;
	}
	else
	{
		if (ClassifyPoint(plane, p[1]) != NEGATIVE)
		{
			// near right is left of left clipping plane

			if (ClassifyPoint(plane, p[2]) != NEGATIVE)
			{
				// far right is left of left clipping plane

				if (ClassifyPoint(plane, p[3]) != NEGATIVE)
				{
					// far left is left of left clipping plane
					return FULL;
				}
			}
		}
		return PARTIAL;
	}

	/* this should never happen */
	ASSERT(0, "unreachable code reached");
	return FULL;
}

Visibility isVisibleQuad(const Plane cull_planes[NUM_PLANES], const Point p[4])
{
	Visibility vis, new_vis;

	vis = FULL;

	new_vis = checkPlaneSide(cull_planes[RIGHT], p);
	if (new_vis == NONE)
	{
		return NONE;
	}
	else if (new_vis == PARTIAL)
	{
		vis = PARTIAL;
	}

	new_vis = checkPlaneSide(cull_planes[LEFT], p);
	if (new_vis == NONE)
	{
		return NONE;
	}
	else if (new_vis == PARTIAL)
	{
		vis = PARTIAL;
	}

	new_vis = checkPlaneSide(cull_planes[FARPLANE], p);
	if (new_vis == NONE)
	{
		return NONE;
	}
	else if (new_vis == PARTIAL)
	{
		vis = PARTIAL;
	}

	new_vis = checkPlaneSide(cull_planes[NEARPLANE], p);
	if (new_vis == NONE)
	{
		return NONE;
	}
	else if (new_vis == PARTIAL)
	{
		vis = PARTIAL;
	}

	return vis;
}

// TODO: SSE, SSE2 extensions...
void GetCullingPlanes(Plane cull_planes[NUM_PLANES])
{
	/* extract the 6 clipping planes from the modelview / projection matrizes */
	float comboMatrix[16], modl[16], proj[16];
	
#ifdef NORMALIZE_PLANES
	float t;
#endif

	/* combine (multiply) modelview and projection matrix into comboMatrix*/
	glGetFloatv(GL_PROJECTION_MATRIX, (float *) &proj);
	glGetFloatv(GL_MODELVIEW_MATRIX, (float *) &modl);

	/* Combine the two matrices (multiply projection by modelview) */
	comboMatrix[ 0] = modl[ 0] * proj[ 0] + modl[ 1] * proj[ 4] + modl[ 2] * proj[ 8] + modl[ 3] * proj[12];
	comboMatrix[ 1] = modl[ 0] * proj[ 1] + modl[ 1] * proj[ 5] + modl[ 2] * proj[ 9] + modl[ 3] * proj[13];
	comboMatrix[ 2] = modl[ 0] * proj[ 2] + modl[ 1] * proj[ 6] + modl[ 2] * proj[10] + modl[ 3] * proj[14];
	comboMatrix[ 3] = modl[ 0] * proj[ 3] + modl[ 1] * proj[ 7] + modl[ 2] * proj[11] + modl[ 3] * proj[15];

	comboMatrix[ 4] = modl[ 4] * proj[ 0] + modl[ 5] * proj[ 4] + modl[ 6] * proj[ 8] + modl[ 7] * proj[12];
	comboMatrix[ 5] = modl[ 4] * proj[ 1] + modl[ 5] * proj[ 5] + modl[ 6] * proj[ 9] + modl[ 7] * proj[13];
	comboMatrix[ 6] = modl[ 4] * proj[ 2] + modl[ 5] * proj[ 6] + modl[ 6] * proj[10] + modl[ 7] * proj[14];
	comboMatrix[ 7] = modl[ 4] * proj[ 3] + modl[ 5] * proj[ 7] + modl[ 6] * proj[11] + modl[ 7] * proj[15];

	comboMatrix[ 8] = modl[ 8] * proj[ 0] + modl[ 9] * proj[ 4] + modl[10] * proj[ 8] + modl[11] * proj[12];
	comboMatrix[ 9] = modl[ 8] * proj[ 1] + modl[ 9] * proj[ 5] + modl[10] * proj[ 9] + modl[11] * proj[13];
	comboMatrix[10] = modl[ 8] * proj[ 2] + modl[ 9] * proj[ 6] + modl[10] * proj[10] + modl[11] * proj[14];
	comboMatrix[11] = modl[ 8] * proj[ 3] + modl[ 9] * proj[ 7] + modl[10] * proj[11] + modl[11] * proj[15];

	comboMatrix[12] = modl[12] * proj[ 0] + modl[13] * proj[ 4] + modl[14] * proj[ 8] + modl[15] * proj[12];
	comboMatrix[13] = modl[12] * proj[ 1] + modl[13] * proj[ 5] + modl[14] * proj[ 9] + modl[15] * proj[13];
	comboMatrix[14] = modl[12] * proj[ 2] + modl[13] * proj[ 6] + modl[14] * proj[10] + modl[15] * proj[14];
	comboMatrix[15] = modl[12] * proj[ 3] + modl[13] * proj[ 7] + modl[14] * proj[11] + modl[15] * proj[15];

	// Normalize the plane equations, if requested
	/* Extract the numbers for the RIGHT plane */
	cull_planes[RIGHT].a = comboMatrix[ 3] - comboMatrix[ 0];
	cull_planes[RIGHT].b = comboMatrix[ 7] - comboMatrix[ 4];
	cull_planes[RIGHT].c = comboMatrix[11] - comboMatrix[ 8];
	cull_planes[RIGHT].d = comboMatrix[15] - comboMatrix[12];

	/* Normalize the result */
#ifdef NORMALIZE_PLANES
	t = (float)sqrt( cull_planes[0].a * cull_planes[0].a + cull_planes[0].b * cull_planes[0].b + cull_planes[0].c * cull_planes[0].c );
	cull_planes[RIGHT].a /= t;
	cull_planes[RIGHT].b /= t;
	cull_planes[RIGHT].c /= t;
	cull_planes[RIGHT].d /= t;
#endif

	/* Extract the numbers for the LEFT plane */
	cull_planes[LEFT].a = comboMatrix[ 3] + comboMatrix[ 0];
	cull_planes[LEFT].b = comboMatrix[ 7] + comboMatrix[ 4];
	cull_planes[LEFT].c = comboMatrix[11] + comboMatrix[ 8];
	cull_planes[LEFT].d = comboMatrix[15] + comboMatrix[12];

	/* Normalize the result */
#ifdef NORMALIZE_PLANES
	t = (float)sqrt( cull_planes[1].a * cull_planes[1].a + cull_planes[1].b * cull_planes[1].b + cull_planes[1].c * cull_planes[1].c );
	cull_planes[LEFT].a /= t;
	cull_planes[LEFT].b /= t;
	cull_planes[LEFT].c /= t;
	cull_planes[LEFT].d /= t;
#endif

	/* Extract the BOTTOM plane */
	cull_planes[BOTTOM].a = comboMatrix[ 3] + comboMatrix[ 1];
	cull_planes[BOTTOM].b = comboMatrix[ 7] + comboMatrix[ 5];
	cull_planes[BOTTOM].c = comboMatrix[11] + comboMatrix[ 9];
	cull_planes[BOTTOM].d = comboMatrix[15] + comboMatrix[13];

	/* Normalize the result */
#ifdef NORMALIZE_PLANES
	t = (float)sqrt( cull_planes[2].a * cull_planes[2].a + cull_planes[2].b * cull_planes[2].b + cull_planes[2].c * cull_planes[2].c );
	cull_planes[BOTTOM].a /= t;
	cull_planes[BOTTOM].b /= t;
	cull_planes[BOTTOM].c /= t;
	cull_planes[BOTTOM].d /= t;
#endif

	/* Extract the TOP plane */
	cull_planes[TOP].a = comboMatrix[ 3] - comboMatrix[ 1];
	cull_planes[TOP].b = comboMatrix[ 7] - comboMatrix[ 5];
	cull_planes[TOP].c = comboMatrix[11] - comboMatrix[ 9];
	cull_planes[TOP].d = comboMatrix[15] - comboMatrix[13];

	/* Normalize the result */
#ifdef NORMALIZE_PLANES
    t = (float)sqrt( cull_planes[3].a * cull_planes[3].a + cull_planes[3].b * cull_planes[3].b + cull_planes[3].c * cull_planes[3].c );
	cull_planes[TOP].a /= t;
	cull_planes[TOP].b /= t;
	cull_planes[TOP].c /= t;
	cull_planes[TOP].d /= t;
#endif
	/* Extract the FAR plane */
	cull_planes[FARPLANE].a = comboMatrix[ 3] - comboMatrix[ 2];
	cull_planes[FARPLANE].b = comboMatrix[ 7] - comboMatrix[ 6];
	cull_planes[FARPLANE].c = comboMatrix[11] - comboMatrix[10];
	cull_planes[FARPLANE].d = comboMatrix[15] - comboMatrix[14];

	/* Normalize the result */
#ifdef NORMALIZE_PLANES
	t = (float)sqrt( cull_planes[4].a * cull_planes[4].a + cull_planes[4].b * cull_planes[4].b + cull_planes[4].c * cull_planes[4].c );
	cull_planes[FARPLANE].a /= t;
	cull_planes[FARPLANE].b /= t;
	cull_planes[FARPLANE].c /= t;
	cull_planes[FARPLANE].d /= t;
#endif

	/* Extract the NEAR plane */
	cull_planes[NEARPLANE].a = comboMatrix[ 3] + comboMatrix[ 2];
	cull_planes[NEARPLANE].b = comboMatrix[ 7] + comboMatrix[ 6];
	cull_planes[NEARPLANE].c = comboMatrix[11] + comboMatrix[10];
	cull_planes[NEARPLANE].d = comboMatrix[15] + comboMatrix[14];

	/* Normalize the result */
#ifdef NORMALIZE_PLANES
	t = (float) sqrt( cull_planes[5].a * cull_planes[5].a + cull_planes[5].b * cull_planes[5].b + cull_planes[5].c * cull_planes[5].c );
	cull_planes[NEARPLANE].a /= t;
	cull_planes[NEARPLANE].b /= t;
	cull_planes[NEARPLANE].c /= t;
	cull_planes[NEARPLANE].d /= t;
#endif
}
